# ========================================================================================================
# Assetto Corsa Tyre Buddy
# - written by Becky York - May 2022
# Designed to help optimise tyre pressures, brake bias & camber
# ========================================================================================================
# v0.0.1 alpha
# Inside/middle/outside tyre temps
# Ensure IMO temps are within +/- 10C and that M is ~50% of the difference between I and O
#
# TODO
# add core temp indicator somewhere
# Camber and Pressure recommendations per tyre
# refactor tyres GUI: scale of 0 to 1 & multiply that by window widthXheight
# queue up the colours in the line list so they don't have to be hard coded
# when moving from not braking to braking - record tyre temps
# apps icon (on/off)
# IMO texts
# +/-5°C letters
# optimize for performance?
# BRAKE BIAS
# during braking - record tyre temps
# when finished braking - record tyre temps
# compare front/rear temperature differences between start and end of braking


import ac
import acsys

import os, sys, platform
sysdir=os.path.dirname(__file__)+'/stdlib'
if platform.architecture()[0] == '64bit':
    sysdir=os.path.dirname(__file__)+'/stdlib64'
sys.path.insert(0, sysdir)
os.environ['PATH'] = os.environ['PATH'] + ';.'
from lib.tyrebuddy_sim_info import info
#from string import Template

#AC constants
class CCAC:
    FRONT_LEFT = 0
    FRONT_RIGHT = 1
    REAR_LEFT = 2
    REAR_RIGHT = 3
    TEMP_INNER = 0
    TEMP_MIDDLE = 1
    TEMP_OUTER = 2

CAC = CCAC()

class CTyre:
    Temp_Core = 0
    Temp_IMO = []
    Temp_IM_Diff = 0
    Temp_IM_Diff_Constrained = 0
    Temp_IM_Diff_Over = 0
    Temp_IM_Diff_Percent = 0
    Temp_MO_Diff = 0
    Temp_MO_Diff_Constrained = 0
    Temp_MO_Diff_Over = 0
    Temp_MO_Diff_Percent = 0
    Temp_IMO_Max = 0
    Temp_IMO_Min = 0
    Pressure = 0
    Wear = 0
    TyreLabel = ""

    def __init__(self, TyreLabel):
        #create a new instance of the list for the instance of the class
        #otherwise they use a singleton
        self.Temp_IMO = [0,0,0]
        self.TyreLabel = TyreLabel

    def updateTyre(self, theTyre):
        global info, CAC
        self.Wear = round(info.physics.tyreWear[theTyre],1)
        self.Pressure = round(info.physics.wheelsPressure[theTyre],1)
        self.Temp_Core = round(info.physics.tyreCoreTemperature[theTyre],1)
        self.Temp_IMO[CAC.TEMP_MIDDLE] = round(info.physics.tyreTempM[theTyre],1)
        #for some reason, temp values in/out is swapped on left side tyres
        if (theTyre in [CAC.FRONT_LEFT,CAC.REAR_LEFT]):
            self.Temp_IMO[CAC.TEMP_INNER] = round(info.physics.tyreTempO[theTyre],1)
            self.Temp_IMO[CAC.TEMP_OUTER] = round(info.physics.tyreTempI[theTyre],1)
        else:
            self.Temp_IMO[CAC.TEMP_INNER] = round(info.physics.tyreTempI[theTyre],1)
            self.Temp_IMO[CAC.TEMP_OUTER] = round(info.physics.tyreTempO[theTyre],1)

        #don't use this it's only for fake screenshots!
        #if (theTyre == CAC.FRONT_LEFT):
        #    self.Temp_IMO[CAC.TEMP_INNER] += 5
        #if (theTyre == CAC.FRONT_RIGHT):
        #    self.Temp_IMO[CAC.TEMP_OUTER] -= 5

        self.Temp_IMO_Max = max(self.Temp_IMO)
        self.Temp_IMO_Min = min(self.Temp_IMO)
        self.Temp_IMO_Spread = round(self.Temp_IMO_Max - self.Temp_IMO_Min,1)
        self.Temp_IM_Diff = round(self.Temp_IMO[CAC.TEMP_INNER] - self.Temp_IMO[CAC.TEMP_MIDDLE],1)
        self.Temp_IM_Diff_Constrained = max(min(self.Temp_IM_Diff,5),-5) #constrain to +/- 5°C
        self.Temp_IM_Diff_Over = max(min(abs(self.Temp_IM_Diff - self.Temp_IM_Diff_Constrained),5),0)
        self.Temp_IM_Diff_Percent = round((self.Temp_IM_Diff_Constrained + 5 ) / 10,2) #percent of 10°C
        self.Temp_MO_Diff = round(self.Temp_IMO[CAC.TEMP_OUTER] - self.Temp_IMO[CAC.TEMP_MIDDLE],1)
        self.Temp_MO_Diff_Constrained = max(min(self.Temp_MO_Diff,5),-5) #constrain to +/- 5°C
        self.Temp_MO_Diff_Over = max(min(abs(self.Temp_MO_Diff - self.Temp_MO_Diff_Constrained),5),0)
        self.Temp_MO_Diff_Percent = round((self.Temp_MO_Diff_Constrained + 5 ) / 10,2) #percent of 10°C

class CTelemetry:
    Tyres = [0,0,0,0]
    Tyres[CAC.FRONT_LEFT] = CTyre('FL')
    Tyres[CAC.FRONT_RIGHT] = CTyre('FR')
    Tyres[CAC.REAR_LEFT] = CTyre('RL')
    Tyres[CAC.REAR_RIGHT] = CTyre('RR')

    #TyreText = Template("$Tyre psi:$psi Core:$C IMOHS:$I/$M/$O/$H/$S Wear:$W")

    #def formatTyre(self, theTyre):
    #    global CAC
    #    return self.TyreText.substitute(
    #          Tyre = self.Tyres[theTyre].TyreLabel
    #        , psi = self.Tyres[theTyre].Pressure
    #        , W = self.Tyres[theTyre].Wear
    #        , C = self.Tyres[theTyre].Temp_Core
    #        , I = self.Tyres[theTyre].Temp_IMO[CAC.TEMP_INNER]
    #        , M = self.Tyres[theTyre].Temp_IMO[CAC.TEMP_MIDDLE]
    #        , O = self.Tyres[theTyre].Temp_IMO[CAC.TEMP_OUTER]
    #        , H = self.Tyres[theTyre].Temp_IMO_Max
    #        , S = self.Tyres[theTyre].Temp_IMO_Spread)

    def update(self):
        global CAC
        self.Tyres[CAC.FRONT_LEFT].updateTyre(CAC.FRONT_LEFT)
        self.Tyres[CAC.FRONT_RIGHT].updateTyre(CAC.FRONT_RIGHT)
        self.Tyres[CAC.REAR_LEFT].updateTyre(CAC.REAR_LEFT)
        self.Tyres[CAC.REAR_RIGHT].updateTyre(CAC.REAR_RIGHT)

class COTData:
    Title = "Tyre Buddy"
    loaded = 0
    shutdown = 0
    deltaT = 0

    Telemetry = CTelemetry()

    mainWindow = 0
    mainWindow_Label = ""

    def startup(self):
        self.createWindow()
        self.loaded = 1

    def shutdown(self):
        self.shutdown = 1

    def update(self, deltaT):
        global CAC
        self.deltaT = deltaT
        self.Telemetry.update()

    def createWindow(self):
        self.mainWindow = ac.newApp(self.Title)
        ac.setSize(self.mainWindow, 230, 130)
        ac.setIconPosition(self.mainWindow, 0, -10000)
        ac.setBackgroundOpacity(self.mainWindow, 0)
        ac.setTitle(self.mainWindow,"")
        #ac.setTitle(self.mainWindow, self.Title)

        #self.mainWindow_Label = ac.addLabel(self.mainWindow,"{debug}\r\n{text}".format(debug="debug",text="text"))
        #ac.setPosition(self.mainWindow_Label, 250, 5)

        #ac.setBackgroundTexture(self.mainWindow, "apps/python/OptimiseTyres/res/main_background.png")

    def MakeLine(self,from_x,from_y,to_x,to_y):
        return [
              [from_x , from_y]
            , [to_x   , to_y  ]
        ]

    #when Diff_Over is > 2.5 then gradient from yellow to red
    #when Diff_Over is < 2.5 then gradient from green to yellow
    #whet Diff_Over is 0; then gradient green based on Diff_Percent
    #whet Diff_Over is 0; then gradient blue based on inverse of Diff_Percent
    def getLineColour(self,Diff_Percent,Diff_Over):
        real_Percent = (abs(Diff_Percent-0.5)*2) ## convert +/-0.5 to 0->1
        ## bias the colours a little bit upwards
        fake_Percent = min(real_Percent + 0.2,1) ## maximum green at 80%
        fake_Over = min(real_Percent - 0.8 + Diff_Over,5) if real_Percent >= 0.8 else 0 ## start adding over_camber at 80%
        red = min(fake_Over,2.5) / 2.5 if fake_Over > 0 else 0
        green = (2.5 - (fake_Over-2.5)) / 2.5 if fake_Over > 2.5 else fake_Percent
        blue = 0 if fake_Over > 0 else abs(fake_Percent-1)
        alpha = 1
        return [red,green,blue,alpha]

    def draw(self):
        global CAC

        lTelemetry = self.Telemetry

        ### this shit needs to be refactored
        # refactor for tyres
        # 0 - 0.5 - 1 (left to right)
        # 0 - 0.5 - 1 (top to bottom)
        # then multiply these numbers by width & height
        # then add to padding location
        BoxHeight = 50
        BoxWidth = 100
        PaddingFront = 10
        PaddingRear = 10
        PaddingLeft = 10
        PaddingRight = 10

        #and this
        Front_LineTop = BoxHeight + PaddingFront
        Front_LineBottom = PaddingFront
        Front_LineMiddle = round( BoxHeight / 2 ,0) + PaddingFront
        Rear_LineTop = BoxHeight + Front_LineTop + PaddingRear
        Rear_LineBottom = Front_LineTop + PaddingRear
        Rear_LineMiddle = round( BoxHeight / 2 ,0) + Front_LineTop + PaddingRear
        Left_LineLeft = PaddingLeft
        Left_LineMiddle = round( BoxWidth / 2 ,0) + PaddingLeft
        Left_LineRight = BoxWidth + PaddingLeft
        Right_LineLeft = Left_LineRight + PaddingRight
        Right_LineMiddle = round( BoxWidth / 2 ,0) + Left_LineRight + PaddingRight
        Right_LineRight = BoxWidth + Left_LineRight + PaddingRight

        makeLines = []

        # and all these
        #left front tyre
        makeLines.append(self.MakeLine(Left_LineLeft,Front_LineTop,Left_LineRight,Front_LineTop))
        makeLines.append(self.MakeLine(Left_LineRight,Front_LineTop,Left_LineRight,Front_LineBottom))
        makeLines.append(self.MakeLine(Left_LineRight,Front_LineBottom,Left_LineLeft,Front_LineBottom))
        makeLines.append(self.MakeLine(Left_LineLeft,Front_LineBottom,Left_LineLeft,Front_LineTop))

        #right front tyre
        makeLines.append(self.MakeLine(Right_LineLeft,Front_LineTop,Right_LineRight,Front_LineTop))
        makeLines.append(self.MakeLine(Right_LineRight,Front_LineTop,Right_LineRight,Front_LineBottom))
        makeLines.append(self.MakeLine(Right_LineRight,Front_LineBottom,Right_LineLeft,Front_LineBottom))
        makeLines.append(self.MakeLine(Right_LineLeft,Front_LineBottom,Right_LineLeft,Front_LineTop))

        #left rear tyre
        makeLines.append(self.MakeLine(Left_LineLeft,Rear_LineTop,Left_LineRight,Rear_LineTop))
        makeLines.append(self.MakeLine(Left_LineRight,Rear_LineTop,Left_LineRight,Rear_LineBottom))
        makeLines.append(self.MakeLine(Left_LineRight,Rear_LineBottom,Left_LineLeft,Rear_LineBottom))
        makeLines.append(self.MakeLine(Left_LineLeft,Rear_LineBottom,Left_LineLeft,Rear_LineTop))

        #right rear tyre
        makeLines.append(self.MakeLine(Right_LineLeft,Rear_LineTop,Right_LineRight,Rear_LineTop))
        makeLines.append(self.MakeLine(Right_LineRight,Rear_LineTop,Right_LineRight,Rear_LineBottom))
        makeLines.append(self.MakeLine(Right_LineRight,Rear_LineBottom,Right_LineLeft,Rear_LineBottom))
        makeLines.append(self.MakeLine(Right_LineLeft,Rear_LineBottom,Right_LineLeft,Rear_LineTop))

        #middle vertical lines (grey)
        makeLines.append(self.MakeLine(Left_LineMiddle,Front_LineBottom,Left_LineMiddle,Front_LineTop))
        makeLines.append(self.MakeLine(Right_LineMiddle,Front_LineBottom,Right_LineMiddle,Front_LineTop))
        makeLines.append(self.MakeLine(Left_LineMiddle,Rear_LineBottom,Left_LineMiddle,Rear_LineTop))
        makeLines.append(self.MakeLine(Right_LineMiddle,Rear_LineBottom,Right_LineMiddle,Rear_LineTop))

        #middle horizontal lines (grey)
        makeLines.append(self.MakeLine(Left_LineLeft,Front_LineMiddle,Left_LineRight,Front_LineMiddle))
        makeLines.append(self.MakeLine(Right_LineLeft,Front_LineMiddle,Right_LineRight,Front_LineMiddle))
        makeLines.append(self.MakeLine(Left_LineLeft,Rear_LineMiddle,Left_LineRight,Rear_LineMiddle))
        makeLines.append(self.MakeLine(Right_LineLeft,Rear_LineMiddle,Right_LineRight,Rear_LineMiddle))


        FL_IM_Diff_Percent = ( lTelemetry.Tyres[CAC.FRONT_LEFT].Temp_IM_Diff_Percent -1 ) * -1 # invert for the UI
        FL_MO_Diff_Percent = ( lTelemetry.Tyres[CAC.FRONT_LEFT].Temp_MO_Diff_Percent -1 ) * -1 # invert for the UI

        FR_IM_Diff_Percent = ( lTelemetry.Tyres[CAC.FRONT_RIGHT].Temp_IM_Diff_Percent -1 ) * -1 # invert for the UI
        FR_MO_Diff_Percent = ( lTelemetry.Tyres[CAC.FRONT_RIGHT].Temp_MO_Diff_Percent -1 ) * -1 # invert for the UI

        RL_IM_Diff_Percent = ( lTelemetry.Tyres[CAC.REAR_LEFT].Temp_IM_Diff_Percent -1 ) * -1 # invert for the UI
        RL_MO_Diff_Percent = ( lTelemetry.Tyres[CAC.REAR_LEFT].Temp_MO_Diff_Percent -1 ) * -1 # invert for the UI

        RR_IM_Diff_Percent = ( lTelemetry.Tyres[CAC.REAR_RIGHT].Temp_IM_Diff_Percent -1 ) * -1 # invert for the UI
        RR_MO_Diff_Percent = ( lTelemetry.Tyres[CAC.REAR_RIGHT].Temp_MO_Diff_Percent -1 ) * -1 # invert for the UI

        #Front
        FL_IM_Diff_Top = round((Front_LineTop - Front_LineBottom) * FL_IM_Diff_Percent ,0) + Front_LineBottom
        FL_MO_Diff_Top = round((Front_LineTop - Front_LineBottom) * FL_MO_Diff_Percent ,0) + Front_LineBottom
        FR_IM_Diff_Top = round((Front_LineTop - Front_LineBottom) * FR_IM_Diff_Percent ,0) + Front_LineBottom
        FR_MO_Diff_Top = round((Front_LineTop - Front_LineBottom) * FR_MO_Diff_Percent ,0) + Front_LineBottom

        #Rear
        RL_IM_Diff_Top = round((Rear_LineTop - Rear_LineBottom) * RL_IM_Diff_Percent ,0) + Rear_LineBottom
        RL_MO_Diff_Top = round((Rear_LineTop - Rear_LineBottom) * RL_MO_Diff_Percent ,0) + Rear_LineBottom
        RR_IM_Diff_Top = round((Rear_LineTop - Rear_LineBottom) * RR_IM_Diff_Percent ,0) + Rear_LineBottom
        RR_MO_Diff_Top = round((Rear_LineTop - Rear_LineBottom) * RR_MO_Diff_Percent ,0) + Rear_LineBottom

        #MidPointLines
        makeLines.append(self.MakeLine(Left_LineLeft,FL_MO_Diff_Top,Left_LineRight,FL_IM_Diff_Top))
        makeLines.append(self.MakeLine(Right_LineLeft,FR_IM_Diff_Top,Right_LineRight,FR_MO_Diff_Top))
        makeLines.append(self.MakeLine(Left_LineLeft,RL_MO_Diff_Top,Left_LineRight,RL_IM_Diff_Top))
        makeLines.append(self.MakeLine(Right_LineLeft,RR_IM_Diff_Top,Right_LineRight,RR_MO_Diff_Top))

        #FL
        makeLines.append(self.MakeLine(Left_LineRight,FL_IM_Diff_Top,Left_LineMiddle,Front_LineMiddle))
        makeLines.append(self.MakeLine(Left_LineLeft,FL_MO_Diff_Top,Left_LineMiddle,Front_LineMiddle))
        #FR
        makeLines.append(self.MakeLine(Right_LineRight,FR_MO_Diff_Top,Right_LineMiddle,Front_LineMiddle))
        makeLines.append(self.MakeLine(Right_LineLeft,FR_IM_Diff_Top,Right_LineMiddle,Front_LineMiddle))
        #RL
        makeLines.append(self.MakeLine(Left_LineRight,RL_IM_Diff_Top,Left_LineMiddle,Rear_LineMiddle))
        makeLines.append(self.MakeLine(Left_LineLeft,RL_MO_Diff_Top,Left_LineMiddle,Rear_LineMiddle))
        #RR
        makeLines.append(self.MakeLine(Right_LineRight,RR_MO_Diff_Top,Right_LineMiddle,Rear_LineMiddle))
        makeLines.append(self.MakeLine(Right_LineLeft,RR_IM_Diff_Top,Right_LineMiddle,Rear_LineMiddle))


        lFL_Tyre = lTelemetry.Tyres[CAC.FRONT_LEFT]
        FL_Colour_IM = self.getLineColour(lFL_Tyre.Temp_IM_Diff_Percent,lFL_Tyre.Temp_IM_Diff_Over)
        FL_Colour_MO = self.getLineColour(lFL_Tyre.Temp_MO_Diff_Percent,lFL_Tyre.Temp_MO_Diff_Over)
        lFR_Tyre = lTelemetry.Tyres[CAC.FRONT_RIGHT]
        FR_Colour_IM = self.getLineColour(lFR_Tyre.Temp_IM_Diff_Percent,lFR_Tyre.Temp_IM_Diff_Over)
        FR_Colour_MO = self.getLineColour(lFR_Tyre.Temp_MO_Diff_Percent,lFR_Tyre.Temp_MO_Diff_Over)
        lRL_Tyre = lTelemetry.Tyres[CAC.REAR_LEFT]
        RL_Colour_IM = self.getLineColour(lRL_Tyre.Temp_IM_Diff_Percent,lRL_Tyre.Temp_IM_Diff_Over)
        RL_Colour_MO = self.getLineColour(lRL_Tyre.Temp_MO_Diff_Percent,lRL_Tyre.Temp_MO_Diff_Over)
        lRR_Tyre = lTelemetry.Tyres[CAC.REAR_RIGHT]
        RR_Colour_IM = self.getLineColour(lRR_Tyre.Temp_IM_Diff_Percent,lRR_Tyre.Temp_IM_Diff_Over)
        RR_Colour_MO = self.getLineColour(lRR_Tyre.Temp_MO_Diff_Percent,lRR_Tyre.Temp_MO_Diff_Over)

        #ac.setText(self.mainWindow_Label,'Percent={} Over={}\r\nr={} g={} b={} a={}'.format(lFL_Tyre.Temp_IM_Diff_Percent,lFL_Tyre.Temp_IM_Diff_Over,round(FL_Colour[0],2),round(FL_Colour[1],2),round(FL_Colour[2],2),round(FL_Colour[3],2)))

        ac.glColor4f(1,1,1,1) #white boxes
        ac.glBegin(0)
        nIndex = 0

        for theLine in makeLines:
            #all this hard coded shit needs to be refactored
            nTyreIndex = 29

            nIndex += 1
            if nIndex % 4 == 0:
                ac.glEnd()
                ac.glBegin(0)
            if nIndex == 17: #grey middle lines
                ac.glColor4f(0.25,0.25,0.25,1)
            if nIndex == 25: #mid point lines
                ac.glColor4f(1,0.6,0.8,0.5)
            if nIndex == nTyreIndex: #left front IM
                ac.glColor4f(*FL_Colour_IM)
            if nIndex == nTyreIndex+1: #left front MO
                ac.glColor4f(*FL_Colour_MO)
            if nIndex == nTyreIndex+2: #right front MO
                ac.glColor4f(*FR_Colour_MO)
            if nIndex == nTyreIndex+3: #right front IM
                ac.glColor4f(*FR_Colour_IM)
            if nIndex == nTyreIndex+4: #rear left IM
                ac.glColor4f(*RL_Colour_IM)
            if nIndex == nTyreIndex+5: #rear left MO
                ac.glColor4f(*RL_Colour_MO)
            if nIndex == nTyreIndex+6: #rear right MO
                ac.glColor4f(*RR_Colour_MO)
            if nIndex == nTyreIndex+7: #rear right IM
                ac.glColor4f(*RR_Colour_IM)
            ac.glVertex2f(theLine[0][0],theLine[0][1])
            ac.glVertex2f(theLine[1][0],theLine[1][1])
        ac.glEnd()

OTData = COTData()

def acMain(ac_version):
    global OTData
    OTData.startup()

    ac.addRenderCallback(OTData.mainWindow , acDoDraw)
    return OTData.Title

def acDoDraw(deltaT):
    #add update limits to reduce cpu cycles
    global OTData
    OTData.draw()

def acUpdate(deltaT):
    #add update limits to reduce cpu cycles
    global OTData
    OTData.update(deltaT)

def acShutdown(*args):
    global OTData
    OTData.shutdown()
